#!/bin/env python3
# coding:utf-8
"""
    为zabbix每个主机组创建用户组,
    zabbix用户组及其成员信息同步至lad中。
"""

import argparse
import logging
import ldap3
from zabbix_api import ZabbixAPIException
from lib.utils.LdapTool import Ldap

parser = argparse.ArgumentParser()
parser.add_argument('--create-ldap-group', action='store_true')
parser.add_argument('--create-zbx-usrgrp', action='store_true')
parser.add_argument('--clean', action='store_true')
parser.add_argument('-l', '--ldap-server', required=True, help='ldap server ip address')
parser.add_argument('-o', '--ldap-port', default=389, help='ldap server port')
parser.add_argument('-b', '--ldap-user', required=True, help='ldap bind user')
parser.add_argument('-w', '--ldap-password', required=True, help='ldap password')
parser.set_defaults(handler=lambda args: main(args))

# ldap搜索dn
LDAP_USER_GROUP_DN = 'ou=user groups,ou=zabbix,dc=shchinafortune,dc=local'
LDAP_USER_DN = 'ou=Person,dc=shchinafortune,dc=local'

def main(args):

    zapi = args.zapi
    ldap = Ldap(args.ldap_server, args.ldap_user, args.ldap_password, args.ldap_port)

    # 1.取zbx主机组
    zbx_hostgrps = zapi.hostgroup.get({
        'output': ['name'],
        'filter': {'flag': 0, 'internal': 0},
        'monitored_hosts': 1
    })

    # 2.取zbx用户组
    zbx_usergrps = zapi.usergroup.get({
        'searchWildcardsEnabled': True,
        'output': ['name'],
        'selectUsers': ['userid', 'alias', 'name', 'surname']
    })

    # 3.筛选出zbx中没有用户组的主机组
    usergrp_names = [usergrp['name'].replace(' admins', '') for usergrp in zbx_usergrps]
    zbx_hostgrps_without_usergrp = filter(lambda g: g['name'] not in usergrp_names, zbx_hostgrps)

    # 4.为上一步筛出的zbx主机组创建用户组
    if args.create_zbx_usrgrp:

        for hostgrp in zbx_hostgrps_without_usergrp:
            try:
                grp_name = hostgrp['name'] + ' admins'

                res = zapi.usergroup.create({
                    'name': grp_name,
                    'rights': {'id': hostgrp['groupid'], 'permission': 3}
                })

                # 同时追加至zbx_usergrps
                zbx_usergrps.append({'usrgrpid': ''.join(res['usrgrpids']), 'name': grp_name, 'users': []})
                logging.info(f'Create zabbix userGroup-> `{grp_name}` success!')

            except ZabbixAPIException as err:
                logging.error(err)


    # 清除ldap中zabbix用户组
    if args.clean:
        logging.info('clean ldap user group `{}`'.format(LDAP_USER_GROUP_DN))
        ldap.clean_usergrp(LDAP_USER_GROUP_DN)

    # 5.更新/新建ldap group
    if args.create_ldap_group:

        logging.info('update ldap user group `{}`'.format(LDAP_USER_GROUP_DN))
        for zbx_grp in zbx_usergrps:

            # 新的成员列表
            uniqueMember = []

            # 判断zabbix用户是否存在于ldap中, 添加至成员列表
            for user in zbx_grp['users']:
                res = ldap.search_user(LDAP_USER_DN, user['alias'])
                if res:
                    uniqueMember.append(res)


            zbx_grp_name  = zbx_grp['name'].replace('(', '\\28').replace(')', '\\29') # 特殊字符需进行处理

            search_res = ldap.search_usergrp(LDAP_USER_GROUP_DN, zbx_grp_name)

            if search_res:
                ldap.update_member(search_res[0], uniqueMember)
            else:
                cn = 'cn=%s,%s' % (zbx_grp_name, LDAP_USER_GROUP_DN)
                ldap.create_usergrp(cn, uniqueMember)




#!/bin/env python3
# coding:utf-8
"""
    从企业微信中获取用户ID，更新到zabbix用户的企业微信告警媒介的sendto
"""

import sys
import argparse
import logging
import os
from functools import reduce
from lib.utils.wework import SyncWeworkMedia


parser = argparse.ArgumentParser()
parser.add_argument('--corpid', help='企业微信的企业ID')
parser.add_argument('--secret', help='企业微信内应用的Secret')
parser.add_argument('-t', '--token-cache', help='从本地文件读取access_token')
parser.add_argument('-d', '--depart-name', help='指定企业微信中部门名称')
parser.add_argument('-e', '--env', action='store_true', help='从环境变量中读取参数')
parser.add_argument('-g', '--usergroups', nargs='+', action='append', help='指定更新的zabbix用户组')
parser.add_argument('--username', help='指定更新的zabbix用户')
parser.add_argument('--allow-update', action='store_true', help='当zabbix user已存在企业微信告警媒介, \
但sendto字段与获取的企业微信userid不一致, 是否允许更新')
parser.set_defaults(handler=lambda args: main(args))


def main(args):
    # print(args)
    corpid = args.corpid
    secret = args.secret
    token = None

    # 从环境变量中读取参数, 优先级小于命令行指定的参数
    if args.env:
        corpid = corpid if corpid else os.environ.get('WEWORK_CORPID')
        secret = secret if secret else os.environ.get('WEWORK_SECRET')
        token = os.environ.get('WEWORK_TOKEN')

    if args.token_cache:
        try:
            with open(args.token_cache, mode='r', encoding='utf-8') as f:
                token = f.readline()
        except Exception as e:
            logging.warning(e)

    if corpid and secret:

        corpapi = SyncWeworkMedia.create_corpapi(corpid, secret, token)

        zbx_user_group_ids = reduce(lambda x, y:x + y ,args.usergroups) if args.usergroups else []

        worker = SyncWeworkMedia(zapi=args.zapi,
                        corpapi=corpapi,
                        depart_name=args.depart_name,
                        zbx_usrgrps=zbx_user_group_ids,
                        zbx_username=args.username,
                        allow_update=args.allow_update)
        worker.run()
        worker.show()
    else:
        parser.print_help()
        logging.error('Missing key parameters `corpid` and `secret`')
        sys.exit(1)



"""
提供模板的显示、添加、清除、替换等操作
"""
from .TemplateBaseTool import BaseTool

class ListTemplate(BaseTool):
    """仅提供格式化输出功能"""
    def __init__(self, args):
        super(ListTemplate, self).__init__(args)
        self.pretty_print()


class AddTemplate(BaseTool):
    """提供添加模板的功能"""
    def __init__(self, args):
        self.templates = args.add
        super(AddTemplate, self).__init__(args)
        self.add()

    def add(self) -> None:
        for h in self.hosts_info:
            params = {'hostid': h['hostid'], 'templates': self.templates_info + h['parentTemplates']}
            self.update_host(params)
        # 打印更新结果
        self.gen_hosts_info()
        self.pretty_print()


class ClearTemplate(BaseTool):
    """移除指定模板"""
    def __init__(self, args):
        self.templates = args.clear
        super(ClearTemplate, self).__init__(args)
        self.clear()

    def clear(self) -> None:
        for h in self.hosts_info:
            params = {'hostid': h['hostid'], 'templates_clear': self.templates_info}
            self.update_host(params)
        # 打印更新结果
        self.gen_hosts_info()
        self.pretty_print()


class UseTemplate(BaseTool):
    """替换掉全部模板"""
    def __init__(self, args):
        self.templates = args.use
        super(UseTemplate, self).__init__(args)
        self.use()

    def use(self) -> None:
        for h in self.hosts_info:
            params = {'hostid': h['hostid'], 'templates': self.templates_info}
            self.update_host(params)
        # 打印更新结果
        self.gen_hosts_info()
        self.pretty_print()


class ReplaceTemplate(BaseTool):
    """替换掉指定模板"""
    def __init__(self, args):
        self.templates = args.replace
        self.instead_templates = args.to
        super(ReplaceTemplate, self).__init__(args)
        self.replace()

    def replace(self) -> None:
        # 替换后的templates
        instead_templates_info = self._get_templates_info(self.instead_templates)
        for h in self.hosts_info:
            # 过滤掉被替换的templates
            new_templates = list(filter(lambda t: t not in self.templates_info, h['parentTemplates']))
            new_templates += instead_templates_info
            params = {'hostid': h['hostid'], 'templates': new_templates}
            self.update_host(params)
        # 打印更新结果
        self.gen_hosts_info()
        self.pretty_print()





#!/bin/env python3
# coding:utf-8
"""
    从企业微信中获取用户ID，更新到zabbix用户的企业微信告警媒介的sendto
"""

import sys
import logging
import copy
from zabbix_api import ZabbixAPIException
from weworkapi.CorpApi import CORP_API_TYPE, CorpApi
from weworkapi.AbstractApi import ApiException
import prettytable as pt

# 告警级别
NOT_CLASSIFIED = 1
INFORMATION = 2
WARNING = 4
AVERAGE = 8
HIGH = 16
DISASTER = 32

# wework告警级别
WEWORK_SEVERITY = [WARNING, AVERAGE, HIGH, DISASTER]
# wework告警启用状态，'0'->enabled, '1'->disabled
WEWORK_ACTIVE = '0'
# wework告警启用时间
WEWORK_PERIOD = '1-7,00:00-24:00'

# 告警媒介类型和属性对照, 用来判断sendto是邮箱|手机号码|(企业)微信用户
MediaAttrList = [
   {"kind": "email", "media_type_name": ["Email"], "attr": "email"},
   {"kind": "sms", "media_type_name": ["FortuneSMS"], "attr": "mobile"},
   {"kind": "wework", "media_type_name": ["wework-sendmsg"], "attr": "wework_id"}
]

# 通过手机号码匹配
MATCH_MOBILE = lambda z_user,w_user: w_user['mobile'] in z_user.get('mobile', [])
# 通过姓名 surname + name 匹配
MATCH_NAME = lambda z_user,w_user: z_user['fullname'] == w_user['name']
# 通过姓名 name + surname 匹配
MATCH_REVERSE_NAME = lambda z_user,w_user: z_user['fullname_reverse'] == w_user['name']
# 通过邮箱匹配
MATCH_EMAIL = lambda z_user,w_user: w_user['email'] in z_user.get('email', [])

# 匹配函数列表, 列表中的方法依次执行, 直到表达式返回True
MATCH_FUNCS = [MATCH_MOBILE, MATCH_NAME, MATCH_REVERSE_NAME, MATCH_EMAIL]



class SyncWeworkMedia:
    """
       1.从zabbix用户告警媒介中，通过报警媒介类型, 提取到用户的手机号码、邮箱、姓名等信息.
       2.通过多种途径匹配到该用户在企业微信的userid.
       3.优先通过手机号码匹配, 如用户无手机号码或手机号码匹配不到，再依次通过其他途径匹配.
       4.最终匹配到企业微信的userid的用户, 新建或更新报警媒介.
    """

    def __init__(self, zapi, corpapi, depart_name=None,
                 zbx_usrgrps=None, zbx_username='',
                 allow_update=False):
        self._zapi = zapi
        self._corpapi = corpapi
        self._usrgrps = zbx_usrgrps
        self._depart_name = depart_name
        self._username = zbx_username
        self._allow_update = allow_update
        self.set_default()


    def set_default(self):
        """
            初始化运行所需参数
        """
        self.media_attr = {}
        self.wework_media_type_ids = []
        self.wework_severity = sum(WEWORK_SEVERITY)
        self.wework_active = WEWORK_ACTIVE
        self.wework_period = WEWORK_PERIOD
        self.match_funcs = MATCH_FUNCS


    def run(self):
        """
            执行同步
        """
        self.set_media_attr()
        self.get_zbx_users()
        self.get_wework_users()
        for z_u in self.zbx_users:
            self.add_user_media_wework(z_u, update=self._allow_update)


    def get_wework_users(self):
        """
            获取企业微信中，指定部门的用户信息, 保存至self.wework_users
        """
        # 从企业微信部门列表中找到指定部门->target_depart
        departs = self._wework_request(CORP_API_TYPE['DEPARTMENT_LIST'])
        target_depart = [d['id'] for d in departs['department'] if d['name'] == self._depart_name]

        if target_depart:
            depart_id = str(target_depart[0])
        else:
            logging.error(f'企业微信中未找到部门:{self._depart_name}.')
            sys.exit(1)

        # 获取该部门下所有人员, 包含子部门的人员
        response = self._wework_request(CORP_API_TYPE['USER_LIST'], {
            'department_id': depart_id,
            'fetch_child': '1'
        })
        self.wework_users = response['userlist']


    def match_wework_userid(self, zbx_user):
        """
            匹配zabbix用户对应的企业微信userid
            依次执行 match_funcs 列表中的匹配方法, 直到表达式返回True停止
            返回对应的userid
        """
        for match_func in self.match_funcs:
            result = [u for u in self.wework_users if match_func(zbx_user, u)]
            if result:
                return result[0]['userid']
        return None


    def get_zbx_users(self):
        """
            获取zabbix中用户相关信息
            根据用户告警媒介获取对应的手机号、邮箱、完整姓名、完整姓名反转等属性
        """

        # 获取指定zabbix用户组的groupid
        gids = [g['usrgrpid'] for g in self.get_zbx_usergroup()]

        # 获取指定groupid/全部的zabbix用户, gids为空列表时, 会返回全部zabbix用户
        params = {
            "output": ["alias", "name", "surname"],
            "usrgrpids": gids,
            "selectMedias": ["mediatypeid", "sendto", "active", "severity", "period"],
            "filter": {"alias": self._username}
        }
        self.zbx_users = self._zbx_request('user.get', params)

        # 根据zabbix用户所具有的media, 为用户添加(mobile|email)等属性, 方便进行匹配
        for zbx_user in self.zbx_users:
            # 添加fullname 和 fullname_reverse，即姓+名 和 名+姓
            zbx_user['fullname'] = zbx_user.get('surname','') + zbx_user.get('name','')
            zbx_user['fullname_reverse'] = zbx_user.get('name','') + zbx_user.get('surname','')

            for media in zbx_user['medias']:
                if media['mediatypeid'] in self.media_attr:
                    attr = self.media_attr[media['mediatypeid']]
                    sendto = media['sendto'] if isinstance(media['sendto'], list) else [media['sendto']]
                    if zbx_user.get(attr):
                        zbx_user[attr] += sendto
                    else:
                        zbx_user[attr] = sendto
                else:
                    continue


    def set_media_attr(self):
        """
            设置 告警媒介类型 与 用户属性(手机号|邮箱|wework_id) 的对应关系
        """
        mts = self._zbx_request('mediatype.get', {
            "output": ["description"]
        })
        # 遍历zabbix所有媒介类型, 根据对照关系, 判断是那种属性(mobile|email|wework_id)
        # 当zabbux用户具有该类型的媒介时, 则用户应具有对应的属性
        for mt in mts:
            for item in MediaAttrList:
                if mt['description'] in item['media_type_name']:
                    self.media_attr[mt['mediatypeid']] = item['attr']
                    # 如果媒介类型是属于企业微信的, 则保存至wework_media_type_ids
                    if item['kind'] == 'wework':
                        self.wework_media_type_ids.append(mt['mediatypeid'])
                    break


    def get_zbx_usergroup(self):
        """
            获取zabbix中用户组名称对应groupid
        """
        return self._zbx_request('usergroup.get', {
            "output": ["name"],
            "filter": {
                "name": self._usrgrps
            }
        })


    def add_user_media_wework(self, zbx_user, update=False):
        """
            为zabbix用户添加企业微信告警媒介。

            update: 如用户已经存在企业微信告警媒介, 且原userid与获取到的userid不一致,
                    值为False 则不做处理，
                    值为True 则更新为获取到的userid。
        """
        wework_userid = self.match_wework_userid(zbx_user)
        if not wework_userid:
            logging.info(f"同步失败: zabbix user {zbx_user['alias']} 未找到对应的企业微信账号.")
            return

        zbx_user_medias = zbx_user['medias']
        zbx_user_medias_copy = copy.deepcopy(zbx_user['medias'])

        for typeid in self.wework_media_type_ids:
            # zabbix user 已经有wework报警媒介
            origin_wework_media = [m for m in zbx_user_medias if m['mediatypeid'] == typeid]
            if origin_wework_media:
                # update为True 则更新为获取到的userid
                [m.update({'sendto': wework_userid}) for m in origin_wework_media if update]
            else:
                # 添加wework报警媒介
                zbx_user_medias.append({
                    'mediatypeid': typeid,
                    'sendto': wework_userid,
                    'active': self.wework_active,
                    'severity': self.wework_severity,
                    'period': self.wework_period
                })

        # 如果zabbix user的media有被修改过, 则进行更新
        if zbx_user_medias != zbx_user_medias_copy:
            self._zbx_request('user.update', {
                'userid': zbx_user['userid'],
                'user_medias': zbx_user['medias']
            })
            zbx_user['wework_id'] = [wework_userid]
            logging.info(f"同步成功: zbx_user:{zbx_user['alias']} wework_userid:{wework_userid}")


    def _zbx_request(self, method, params):
        """
            封装对zapi的调用
        """
        try:
            module, func = method.split('.')
            return getattr(getattr(self._zapi, module), func)(params)
        except ZabbixAPIException as err:
            logging.error(err)
            sys.exit(1)


    def _wework_request(self, api_type, params=None):
        """
            封装对corpapi的调用
        """
        try:
            return self._corpapi.httpCall(api_type, params)
        except ApiException as err:
            logging.error(err)
            sys.exit(1)


    def show(self):
        """表格化打印"""
        table = pt.PrettyTable()
        table.title = 'Zabbix用户企业微信账号对照'
        table.field_names = ['Zabbix User', 'Full Name', 'Wework UserId']
        table.align['Zabbix User'] = "l"
        table.align['Full Name'] = "l"
        table.align['Wework UserId'] = "l"

        for zbx_user in self.zbx_users:
            table.add_row([zbx_user['alias'], zbx_user['fullname'], zbx_user.get('wework_id', [])])
        print(table)


    @staticmethod
    def create_corpapi(corpid:str, secret:str, token:str=None) ->CorpApi:
        """
            创建企业微信api实例
        """
        corpapi = CorpApi(corpid, secret)
        corpapi.access_token = token
        return corpapi




#!/bin/env python3
# coding:utf-8
"""
    Mgnt Servers主机组内主机与其他主机组做序列号匹配,
    匹配上的把普通主机的inventory OOB IP address的值
    设置为 Mgnt Servers主机 snmp interface的地址
"""

import argparse
import logging
from zabbix_api import ZabbixAPIException

from lib.utils.inventory_tag import InventoryTagDict

parser = argparse.ArgumentParser("Matching inventory OOB IP address")
# parser.add_argument('-g', '--group_name', help='Specify a group name')
parser.add_argument('--rm_auto_oob', action='store_true',
    help='Remove auto_oob in inventory tag field and reset the oob_ip inventory field')
parser.add_argument('--rm_auto_server', action='store_true',
    help='Remove auto_server=x.x.x.x in inventory tag field')
parser.add_argument('--tags', action='store_true',
    help='Make server and bmc host inventory tag')
parser.set_defaults(handler=lambda args: main(args))


def rm_auto_oob_host_inventory_tag(zapi):
    hosts = zapi.host.get({
        'output': ['hostid'],
        'selectInventory': ['tag', 'oob_ip'],
        'searchInventory': {'tag': 'auto_oob'}
    })
    logging.info(f'{len(hosts)} hosts have auto_oob tag')
    for host in hosts:
        itd = InventoryTagDict(host['inventory']['tag'])
        del itd['auto_oob']
        logging.info(f'Updating {host} inventory tag and oob_ip for remove auto_oob tag and oob_ip')
        zapi.host.update({
            'hostid': host['hostid'], 
            'inventory': {'tag': str(itd), 'oob_ip':'' }
        })

def rm_auto_server_host_inventory_tag(zapi):
    hosts = zapi.host.get({
        'output': ['hostid'],
        'selectInventory': ['tag'],
        'searchInventory': {'tag': 'auto_server'}
    })
    logging.info(f'{len(hosts)} hosts have auto_server tag')
    for host in hosts:
        itd = InventoryTagDict(host['inventory']['tag'])
        del itd['auto_server']
        logging.info(f'Updating {host} inventory tag for remove auto_server tag')
        zapi.host.update({
            'hostid': host['hostid'], 
            'inventory': {'tag': str(itd)}
        })

def get_server_bmc_hosts(zapi):

    def get_hosts_by_inventory_type(inventory_type):
        return list(
            filter(
                lambda x: len(x['inventory']['serialno_a']) and len(x['interfaces']),  
                zapi.host.get({
                    'output': ['hostid', 'name'],
                    'selectInventory': ['tag', 'serialno_a'],
                    'selectInterfaces': ['type', 'main', 'ip'],
                    # 'groupids': [item['groupid'] for item in plain_groups]
                    'searchInventory': {'type': inventory_type},
                })
            )
        )
        # return zapi.host.get({
        #     'output': ['hostid', 'name'],
        #     'selectInventory': ['tag', 'serialno_a'],
        #     'selectInterfaces': ['type', 'main', 'ip'],
        #     # 'groupids': [item['groupid'] for item in plain_groups]
        #     'searchInventory': {'type': inventory_type},
        # })

    return get_hosts_by_inventory_type('Server'), get_hosts_by_inventory_type('BMC')

def handle_hosts_tag(zapi):
    server_hosts, bmc_hosts = get_server_bmc_hosts(zapi)
    logging.info(f'Get validate {len(server_hosts)} Server hosts and {len(bmc_hosts)} BMC hosts')

    server_serials = [ item['inventory']['serialno_a'] for item in server_hosts ]
    bmc_serials = [ item['inventory']['serialno_a'] for item in bmc_hosts ]
    server_serial_dict = dict(zip(server_serials, server_hosts))
    bmc_serial_dict = dict(zip(bmc_serials, bmc_hosts))
    
    match = list(set(server_serials) & set(bmc_serials))
    logging.info(f'The number of {len(match)} serialno matched between Server and BMC hosts')
    for serialno in match:
        server_host = server_serial_dict[serialno]
        bmc_host = bmc_serial_dict[serialno]
        logging.info(f'The serialno_a {serialno} is matched between server {server_host["name"]} and bmc {bmc_host["name"]}')

        # 更新Server 主机清单的tag和oob_ip字段
        if len(bmc_host['interfaces']):
            itd = InventoryTagDict(server_host['inventory']['tag'])
            itd['auto_oob'] = None
            zapi.host.update({
                'hostid': server_host['hostid'],
                'inventory': {
                    'tag': str(itd),
                    'oob_ip': bmc_host['interfaces'][0]['ip']
                }
            })

        # 更新MBC 主机清单的tag字段
        if len(server_host['interfaces']):
            itd = InventoryTagDict(bmc_host['inventory']['tag'])
            itd['auto_server'] = server_host['interfaces'][0]['ip']
            zapi.host.update({
                'hostid': bmc_host['hostid'],
                'inventory': {
                    'tag': str(itd)
                }
            })


def main(args):

    zapi = args.zapi

    # 清除实体服务器inventory tag字段中的auto_oob标识
    if args.rm_auto_oob:
        rm_auto_oob_host_inventory_tag(zapi)

    # 清除BMC的inventory tag字段中的auto_server标识
    if args.rm_auto_server:
        rm_auto_server_host_inventory_tag(zapi)

    # 设置Server的auto_oob和BMC的auto_server inventory tag
    if args.tags:
        handle_hosts_tag(zapi)


   



from unittest import TestCase
from lib.utils.inventory_tag import parse_inventory_tag_literal, InventoryTagDict

class TestInventoryTag(TestCase):

    def test_parse_inventory_tag_literal(self):
        inventory_tag = 'tag1; tag2=val2; =tag3; tag4= '
        dt = parse_inventory_tag_literal(inventory_tag)
        self.assertEqual(dt['tag1'], None)
        self.assertEqual(dt['tag2'], 'val2')
        self.assertNotIn('tag3', dt)
        self.assertEqual(dt['tag4'], '')

    def test_InventoryTagDict(self):
        inventory_tag = 'tag1; tag2=val2; =tag3; tag4= '
        itd = InventoryTagDict(inventory_tag)
        self.assertEqual(itd['tag1'], None)
        self.assertEqual(itd['tag2'], 'val2')
        self.assertNotIn('tag3', itd)
        self.assertEqual(itd['tag4'], '')

        itd['tag5'] = 'val5'
        itd['tag6'] = None
        self.assertEqual(itd['tag5'], 'val5')
        self.assertEqual(itd['tag6'], None)

        # 当value的值不为str或者None时，抛出TypeError
        with self.assertRaises(TypeError):
            itd['tag7'] = ['1', '2']

    def test_InventoryTagDict_str(self):
        itd = InventoryTagDict()
        itd['tag1'] = 'val1'
        itd['tag2'] = None

        self.assertEqual(str(itd), 'tag1=val1;tag2;')

        itd2 = InventoryTagDict()
        itd2['tag;1'] = 'val1'
        itd2['tag2='] = None
        self.assertEqual(str(itd2), r'tag\3B1=val1;tag2\3D;')



import re
import collections


def _safe_tag_literal(ori: str):
    res = ori.replace('=', '\\3D')
    res = res.replace(';', '\\3B')
    return res.strip()


def parse_inventory_tag_literal(inventory_tag):
    res = {}

    if inventory_tag == None:
        return res

    if not isinstance(inventory_tag, str):
            raise TypeError()

    tags = inventory_tag.strip().split(';')

    for item in tags:

        item = item.strip()
        pos = item.find('=')
        if pos < 0:
            key = item.rstrip()
            if len(key) > 0:
                res[key] = None
        elif pos > 0:
            key = item[0: pos]
            val = item[pos+1 : ]
            key = key.rstrip()
            val = val.lstrip()
            if len(key) > 0:
                res[key] = val

    return res

class InventoryTagDict(collections.UserDict):
    """
    由于Zabbix 4.0版本host不支持tag功能，缓解方法是在host inventory的tag字段定稿相应的tag
    inventory tag的格式规范是以分号分隔的k=v或者k字符串，示例如下：
        key1=val1;key2;key3=val3
    """
    def __init__(self, inventory_tag=None):
        super().__init__(parse_inventory_tag_literal(inventory_tag))

    def __setitem__(self, key, item):
        '''
        value的数据类型必须为字符串或者None，不支持其它复杂的数据类型
        '''
        if not isinstance(item, str) and item != None:
            raise TypeError()
        super().__setitem__(key, item)
        
    def __str__(self):
        res = ''
        for k,v in self.data.items():

            res += _safe_tag_literal(k)
            if v:
                res += ('=' + _safe_tag_literal(v))
            res += ';'

        return res


import logging
from zabbix_api import ZabbixAPIException
import prettytable as pt

class BaseTool:
    """提供模板的公共操作"""
    templates = []

    def __init__(self, args):
        self.zapi = args.zapi
        self.hosts = args.hosts or ['']
        self.groups = args.groups or ['']
        self.gen_hosts_info()
        self.gen_templates_info()

    def get_all_hosts(self) -> None:
        """指定的hosts名称 + 指定的group中的hosts名称"""
        if self.groups:
            groups_info = self.zapi.hostgroup.get({
                'filter': {'name': self.groups},
                'selectHosts': ['host']
            })
            self.hosts += [h['host'] for g in groups_info for h in g['hosts']]

    def gen_hosts_info(self) -> None:
        """获取hosts的信息"""
        self.get_all_hosts()
        self.hosts_info = self.zapi.host.get({
            'output': ['host'],
            'selectParentTemplates': ['host'],
            'filter': {'host': self.hosts}
        })

    def _get_templates_info(self, tpl_names: list) -> list:
        """内部方法, 提供的模板名称, 获取对应的模板信息"""
        return self.zapi.template.get({'filter': {'host': tpl_names}, 'output': ['host']})

    def gen_templates_info(self) -> None:
        """获取参数指定的模板信息"""
        self.templates_info = self._get_templates_info(self.templates)

    def update_host(self, params) -> None:
        """更新host"""
        try:
            self.zapi.host.update(params)
            logging.info('update success! hostid: {}'.format(params['hostid']))
        except ZabbixAPIException as err:
            logging.error(err)

    def pretty_print(self) -> None:
        """表格化打印"""
        tb = pt.PrettyTable()
        tb.title = 'Results of linked templates'
        tb.field_names = ['Host', 'Templates']
        tb.align['Host'] = "l"
        tb.align['Templates'] = "l"

        for h in self.hosts_info:
            tb.add_row([h['host'], ', '.join([t['host'] for t in h['parentTemplates']])])
        print(tb)



"""
提供ldap常用的操作
目前支持查找用户, 清除用户组, 创建用户组, 修改用户组成员
"""
import logging
import ldap3
import functools

class Ldap:
    """ldap usergroup 删除、添加、更新"""
    def __init__(self, host, user, passwd, port=389):
        self.logged_in = False
        self.login(ldap3.Server(host, port), user, passwd)

    def login(self, ldap_server, user, passwd):
        """登录"""
        try:
            self._conn = ldap3.Connection(ldap_server, user, passwd, auto_bind=True)
            self.logged_in = True
        except Exception as err:
            logging.error(err)

    def check_login(fun):
        """类内装饰器 检查ldap是否已登录"""
        @functools.wraps(fun)
        def _fun(self, *args, **kwargs):
            if self.logged_in:
                return fun(self, *args, **kwargs)
            else:
                logging.error('no login')
        return _fun

    @check_login
    def search_user(self, dn, user) -> str:
        """查询并返回用户信息"""
        res = self._conn.search(
            search_base = dn,
            search_scope = ldap3.SUBTREE,
            search_filter = '(uid={})'.format(user),
            attributes = ldap3.ALL_ATTRIBUTES
        )
        return self._conn.response[0]['raw_dn'] if res else None

    @check_login
    def search_usergrp(self, dn, usergrp='*') -> list:
        """查询并返回用户组信息"""
        res = self._conn.search(
            search_base = dn,
            search_scope = ldap3.SUBTREE,
            search_filter = '(cn={})'.format(usergrp),
            attributes = ldap3.NO_ATTRIBUTES
        )
        return [g['dn'] for g in self._conn.response]


    @check_login
    def clean_usergrp(self, dn) -> None:
        """清除用户组"""
        for usergrp in self.search_usergrp(dn):
            try:
                self._conn.delete(usergrp)
            except Exception as err:
                logging.error(err)

    @check_login
    def create_usergrp(self, dn, member:list=None) -> None:
        """创建用户组, 添加组成员"""
        try:
            self._conn.add(dn, 'groupOfUniqueNames')
            if member:
                self.update_member(dn, member)
        except Exception as err:
            logging.error(err)

    @check_login
    def update_member(self, cn, member:list=None) -> None:
        if not member:
            member = []
        try:
            self._conn.modify(cn, {'uniqueMember': [(ldap3.MODIFY_REPLACE, member)]})
        except Exception as err:
            logging.error(err)



import sys
from setuptools import setup, find_packages
from setuptools.command.test import test

with open("requirements.txt") as reqs_file:
    reqs = reqs_file.readlines()

class PyTest(test):
    def run_tests(self):
        import pytest
        errno = pytest.main()
        sys.exit(errno)


setup(
    name ="zbxtool-cfomp",
    packages = find_packages(),
    include_package_data=True,
    zip_safe=True,
    install_requires = reqs,
    use_scm_version = {
        "relative_to": __file__,
        "local_scheme": "no-local-version",
    },
    entry_points={
        'console_scripts':[
            'zbxtool = lib.cli:main'
        ],
    },
    setup_requires = ['setuptools_scm==5.0.2']
)



#!/bin/env python3
# coding:utf-8
"""
检索 "Chassis Model" 监控项，如果最新值包含 "VMware" "KVM", 把inventory type字段设置为VM
检索监控项key为proc.num[rsync,root,,--daemon]的主机，如果监控项最新值为1，在inventory tag中添加rsyncd标签，
如果为0，删除在inventory tag中rsyncd标签
"""
import argparse
import logging
from zabbix_api import ZabbixAPIException
from lib.utils.inventory_tag import InventoryTagDict

parser = argparse.ArgumentParser()
parser.set_defaults(handler=lambda args: main(args))

def is_vm(_, item: dict) -> bool:
    """判断字符串是否包含'VMware', 'KVM'"""
    keywords = ['VMware', 'KVM']
    for k in keywords:
        if k in item['lastvalue']:
            return True
    return False

def is_host(item: dict) -> bool:
    """判断该item为host的item, 排除template的item"""
    return item['hosts'][0]['status'] != '3'


class UpdateInventory:
    """更新zabbix inventory"""
    # item.get的参数
    query_params = {}
    # item结果进行过滤的函数
    filter_func = None

    def __init__(self, zapi):
        self.zapi = zapi
        self.items = None
        self.get_items()
        self.update_host()

    def get_items(self):
        """获取items"""
        if self.query_params:
            self.items = self.zapi.item.get(self.query_params)

        # 过滤掉host类型为模板的情况
        self.items = [i for i in self.items if is_host(i)]

        # 执行指定的过滤
        if self.filter_func:
            self.items = [i for i in self.items if self.filter_func(i)]

    def update_host(self):
        """获取以上items的host信息, 更新inventory"""
        for item in self.items:
            hosts_info = self.zapi.host.get({
                'output': ['host'],
                'hostids': [h['hostid'] for h in item['hosts']],
                'selectInventory': "extend"
            })

            # 更新host的inventory
            for host in hosts_info:
                update_params = {'hostid': host['hostid']}
                update_params['inventory'] = self.gen_inventory(item, host)

                # 如inventory mode为禁用, 则改为自动
                if host['inventory']['inventory_mode'] == '-1': # disabled
                    update_params['inventory_mode'] = '1'       # automatic
                try:
                    self.zapi.host.update(update_params)
                    logging.info('update host->%s success!', host['host'])
                except ZabbixAPIException as err:
                    logging.exception("zapi error: %s", err)

    def gen_inventory(self, item:dict, host:None) -> dict:
        raise NotImplementedError("You need to implement method gen_inventory")


class UpdateInventoryType(UpdateInventory):
    """
    检索 "Chassis Model" 监控项，如果最新值包含 "VMware" "KVM", 把inventory type字段设置为VM
    """
    query_params = {
        'output': ['lastvalue'],
        'selectHosts': ['host', 'status'],
        'filter': {'name': 'Chassis Model'}
    }
    filter_func = is_vm

    def gen_inventory(self, item: dict, host: None) -> dict:
        return {'type': 'VM'}


class UpdateInventoryTag(UpdateInventory):
    """
    检索监控项key为proc.num[rsync,root,,--daemon]的主机，如果监控项最新值为1，在inventory tag中添加rsyncd标签，
    如果为0，删除在inventory tag中rsyncd标签
    """
    query_params = {
        'output': ['lastclock', 'lastvalue'],
        'selectHosts': ['host', 'status'],
        'search': {
            'key_': 'proc.num[rsync,root,,--daemon]'
        }
    }

    def gen_inventory(self, item:dict, host:dict) -> dict:
        tags = InventoryTagDict(host['inventory']['tag'])
        # 如果监控项最新值为1，在inventory tag中添加rsyncd标签，
        if item['lastclock'] != '0' and item['lastvalue'] == '1':
            tags['rsyncd'] = '1'
        # 如果监控项最新值为0，删除inventory tag中的rsyncd标签
        elif item['lastclock'] != '0' and item['lastvalue'] == '0':
            if 'rsyncd' in tags:
                del tags['rsyncd']
        return {'tag': str(tags)}


def main(args):
    zapi = args.zapi
    UpdateInventoryType(zapi)
    UpdateInventoryTag(zapi)



#!/bin/env python3
# coding:utf-8
"""
    添加/删除/显示 主机模板
"""

import argparse
import logging
from lib.utils.TemplateTool import *

# 设置子命令和可选参数
parser = argparse.ArgumentParser("(list|add|del) zabbix hosts' templates")
parser.add_argument('--hosts', nargs='+', help="specific zabbix hosts")
parser.add_argument('--groups', nargs='+', help="specific zabbix hostgroups")
parser.add_argument('--to', nargs='+', help='specific templates names instead to')

opt_group = parser.add_mutually_exclusive_group(required=True)
opt_group.add_argument('--list', action='store_true', help="list specific hosts' templates")
opt_group.add_argument('--add', nargs='+', help="add specific hosts' templates")
opt_group.add_argument('--clear', nargs='+', help="del specific hosts' templates")
opt_group.add_argument('--use', nargs='+', help="use specific hosts' templates")
opt_group.add_argument('--replace', nargs='+', help="replace specific hosts' templates")

parser.set_defaults(handler=lambda args:main(args))

def main(args: dict) -> None:
    # 显示模板信息
    if args.list:
        ListTemplate(args)
    # 添加模板
    if args.add:
        AddTemplate(args)
    # 移除指定模板
    if args.clear:
        ClearTemplate(args)
    # 替换全部模板
    if args.use:
        UseTemplate(args)
    # 替换指定模板
    if args.replace:
        if not args.to:
            parser.print_help()
            logging.error('the argument --to is required')
            exit(-1)
        ReplaceTemplate(args)



"""
在各主机上创建总磁盘空间和已用磁盘空间两个监控项。
ZABBIX 3.0版本没有直接获取总磁盘和已用磁盘空间的监控项，只有各挂载的各文件系统的空间使用情况，
因此在各挂载文件系统监控项的基础上创建一个汇总的Calculated监控项
涉及的监控项为 vfs.fs.size[fs,total] vfs.fs.size[fs,used]
创建的计算监控项key为 vfs.fs.totalsize vfs.fs.usedsize
"""
import argparse
import logging


parser = argparse.ArgumentParser()
parser.set_defaults(handler=lambda args: main(args))


APP_NAME = 'Filesystem calculation'

def calculated_formula(items, mode):
    # last("vfs.fs.size[/,total]") + last("vfs.fs.size[/boot,total]")
    filter_keys = ["last(\"" + item["key_"] + "\")" for item in items if mode in item["key_"]]
    formula = '+'.join(filter_keys)
    return formula if formula else "0"


def create_calculated_disk_space_item(api, host, item_name, item_key, params):
    apps = api.application.get({
        'hostids': host["hostid"],
        'filter': {"name": APP_NAME}
    })
    if len(apps) == 0:
        logging.warning("Warning, {0} has no {1} application".format(host['name'], APP_NAME))
    item = {"delay": 3600,
            "hostid": host["hostid"],
            # "interfaceid": [interface["interfaceid"] for interface in host["interfaces"] if interface["type"] == 1],
            "key_": item_key,
            "name": item_name,
            # calculated
            "type": 15,
            # numeric unsigned
            "value_type": 3,
            # decimal
            "data_type": 0,
            "applications": [app["applicationid"] for app in apps],
            "units": "B",
            "params": params
            }

    result = api.item.create(item)
    logging.debug(f"{host['name']} {item['name']} creating success.")
    return result


# def create_aggregate_item(api, hosts, groups, key):
#     agg_item_name = 'Total disk space in $1'
#     agg_item_key = 'grpsum["%s","%s",last]' %

def update_calculated_item_formula(api, items, formula):
    for item in items:
        if item["params"] != formula:
            item["params"] = formula
            api.item.update({
                'itemid': item["itemid"], 
                'params': formula
            })
            logging.info(f"{item} formula is updated.")
        else:
            logging.debug(f"{item} formula is unchanged.")


def upinsert_total_disk_space_item(api, hosts):
    total_item_name = "Total disk space"
    total_item_key = "vfs.fs.totalsize"
    used_item_name = "Used disk space"
    used_item_key = "vfs.fs.usedsize"
    for host in hosts:
        # create_calculated_application(api, host)
        # fs_size_items = api.item.get(hostids=host["hostid"], search={"key_": "vfs.fs.size"})
        fs_size_items = api.item.get({
            'hostids': host["hostid"], 
            'search': {
                "key_": "vfs.fs.size"
                }
        })
        total_formula = calculated_formula(fs_size_items, "total")
        # total_items = api.item.get(hostids=host["hostid"], filter={"name": total_item_name})
        total_items = api.item.get({
            'hostids': host["hostid"], 
            'filter':{
                "name": total_item_name
            }
        })
        if len(total_items) == 0:
            create_calculated_disk_space_item(api, host, total_item_name, total_item_key, total_formula)
        else:
            update_calculated_item_formula(api, total_items, total_formula)

        # used_items = api.item.get(hostids=host["hostid"], filter={"name": used_item_name})
        used_items = api.item.get({
            'hostids': host["hostid"], 
            'filter': {"name": used_item_name}
        })
        used_formula = calculated_formula(fs_size_items, "used")
        if len(used_items) == 0:
            create_calculated_disk_space_item(api, host, used_item_name, used_item_key, used_formula)
        else:
            update_calculated_item_formula(api, used_items, used_formula)


def main(args):
    zapi = args.zapi
    hosts = zapi.host.get({
        'output': ['hostid', 'name'],
        'filter': {
            'available': 1,
            'status': 0
        },
        'searchInventory': {
            'os_short': ['Linux', 'Windows']
        },
        'searchByAny': True
    })
    
    upinsert_total_disk_space_item(zapi, hosts)